/* 
 * cfg.h
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#if STHREADS && !MTCAMP
#include <st.h>
#endif

#ifndef PARALLEL
#define PARALLEL 1
#endif

#ifndef CFG_DBG
#define CFG_DBG 1
#endif

#include <st_list.h>

enum { HEAD_IS_LEFT = 0, HEAD_IS_RIGHT = 1, LEX_HEAD = 2, UNARY_HEAD = 3, };
/* this is never `new'ed */
struct edge
{
  edge() {}
  void * operator new(size_t sz) {
    fprintf(st_errout, "do not new edge\n");
    st_wg_die((void *)1);
  }
  int k_head;			/* head and k */
  int s0_s1;			/* s0 and s1 */
};
typedef edge edge_t;

#define EDGE_S0_S1_S1_WIDTH 16
#define EDGE_S0_S1_S1_MASK ((1 << EDGE_S0_S1_S1_WIDTH) - 1)
#define MK_EDGE_S0_S1(s0, s1) ((s0 << EDGE_S0_S1_S1_WIDTH) + s1)
#define EDGE_K_HEAD_HEAD_WIDTH 2
#define EDGE_K_HEAD_HEAD_MASK 3	/* 2^2 - 1 */
#define MK_EDGE_K_HEAD(k, head) ((k << EDGE_K_HEAD_HEAD_WIDTH) + head)

#define EDGE_K(e) (((e).k_head) >> EDGE_K_HEAD_HEAD_WIDTH)
#define EDGE_HEAD(e) (((e).k_head) & EDGE_K_HEAD_HEAD_MASK)
#define EDGE_S0(e) (((e).s0_s1) >> EDGE_S0_S1_S1_WIDTH)
#define EDGE_S1(e) (((e).s0_s1) & EDGE_S0_S1_S1_MASK)

#define SYMBOL_TABLE_MOTHER_WIDTH 11
enum { 
    bin_rule_size = 760,
    una_rule_size = 1100,
    symbol_max_size = 1100,
};
#define SYMBOL_TABLE_MOTHER_MASK ((1 << SYMBOL_TABLE_MOTHER_WIDTH) -1)

enum istr_status_kind {
    istr_status_none, 
    istr_status_working, 
    istr_status_done
};

template<class T>
struct istr : st_mm
{
  istr() { }

  T val;
  st_int_loc_t status;
#if PARALLEL
  st_join_counter_t j[1];
#endif

  inline void clear(T v) {
    val = v;
    ST_INT_LOC_INIT(&status, istr_status_none);
#if PARALLEL
    st_join_counter_init(j, 1);
#endif
  }
  
  inline T get() {
    if(ST_INT_LOC_CHECK(&status, istr_status_done)) {
      return val;
    } else {
#if PARALLEL
      st_join_counter_wait(j);
#endif
#if CFG_DBG
      st_assert(ST_INT_LOC_CHECK(&status, istr_status_done));
#endif
      return val;
    }
  }

  inline int acquire() {
    if (!ST_INT_LOC_CHECK(&status, istr_status_none)) return 0;
    else {
#if PARALLEL
      int s = st_read_and_lock_int(&status);
      if (s == istr_status_none) {
	st_write_and_unlock_int(&status, istr_status_working); /* unlock */
	return 1;
      } else {
	st_write_and_unlock_int(&status, s); /* unlock */
	return 0;
      }
#else
      status.v = ST_INT_LOC_ATTATCH_TAG(istr_status_working);
      return 1;
#endif
    }
  }

  inline void put(T v) {
#if CFG_DBG
    st_assert(ST_INT_LOC_CHECK(&status, istr_status_working));
#endif
    val = v;
    status.v = ST_INT_LOC_ATTATCH_TAG(istr_status_done);
#if PARALLEL
    st_join_counter_finish(j);
#endif
  }

};

typedef istr<st_list_t(int)> cky_matrix_elem_type;
typedef cky_matrix_elem_type * cky_matrix_t;
typedef st_list_t(edge_t) * result_matrix_t;

struct grammar : st_mm
{
  grammar() {
    binary_rules = 0;
    unary_rules = 0;
  }
  /* register M -> L R */
  void set_binary_rule(int /* M */, int /* L */, int /* R */, 
		       int /* HEAD */, st_region_t);
  /* register M -> C */
  void set_unary_rule(int /* M */, int /* C */, st_region_t);
  void set_lex_rule(int /* M */, int /* C */, st_region_t);
  void feed_grammar(char * /* grammar_file */, st_region_t);
  void read_grammar(char * /* grammar_file */, st_region_t);
  /* for each element s0 in S0
       for each element s1 in S1
         find all rules in grm of the form s -> s0 s1
         and accumulate all S on SYMS 
         also record generated edges in *RES */
  st_list_t(int) 
    apply_binary_rules(st_list_t(int) /* S0 */, st_list_t(int) /* S1 */, 
		       int /* k */, st_list_t(int) /* SYMS */, 
		       st_list_t(edge_t) * /* RES */, st_region_t);

  /* for each element i in I
       find symbols which are transitively generated by applying 
       unary rules from symbol i.
       accumulate them on SYMS.
       also record generated edges in *RES */
  st_list_t(int) 
    apply_unary_rules(st_list_t(int) /* I */, st_list_t(int) /* SYMS */, 
		      st_list_t(edge_t) * /* RES */, st_region_t);

  st_list_t(int) ** binary_rules;
  st_list_t(int) * unary_rules;
};
grammar * make_grammar(char *, st_region_t);

#define CHAR_SPACE ((char)32)
#define CHAR_NEWLINE ((char)10)
int read_token_int(FILE *, char);

#define MAX_CKY_MAT_SIZE 128
#define MREF(m, i, j) m[(i) * MAX_CKY_MAT_SIZE + (j)]
#define MSETF(m, i, j, x) MREF(m, i, j) = (x)

cky_matrix_t make_cky_matrix(st_region_t rg);
void clear_cky_matrix(cky_matrix_t, int);
result_matrix_t make_result_matrix(st_region_t rg);
void clear_result_matrix(result_matrix_t, int);

void cky_parse(st_list_t(int), 
	       grammar *, cky_matrix_t, result_matrix_t, 
	       st_region_t rg);

void show_cky_matrix_1(cky_matrix_t, int, int);
void show_cky_matrix(cky_matrix_t, int);
int count_syms_1(cky_matrix_t, int, int);
int count_syms(cky_matrix_t, int);
void show_count_syms(cky_matrix_t, int);


