/* 
 * qs.c
 */

#ifndef PARALLEL
#define PARALLEL 1
#endif

/* demonstration scenario:
   (1) compile by stgcc --xmtcamp qsm.c
   (2) run: ./a.out 500000 and ./a.out 500000 -nw 10
   (3) uncomment st_begin_profile and st_end_profile below
   (4) compile and run again
   (5) mkxgrph 00stprof 10 > data.xgr
   (6) xgraph data.xgr
 */
#include <strings.h>

#include <st.h>
#include <st_mm.h>

typedef int elem_type;

void print_array(elem_type * A, int a, int b)
{
  int i;
  for (i = a; i <= b; i++) {
    fprintf(stderr, "%d ", A[i]);
    if ((i - a + 1) % 10 == 0) printf("\n");
  }
}

void check_array(elem_type * A, int a, int b, int p, int q, elem_type pivot)
{
  int i;
  fprintf(stderr, "check_array A[%d] ... A[%d], p = %d, q = %d, piv = %d\n",
	  a, b, p, q, pivot);
  print_array(A, a, b);
  if (p <= q) {
    fprintf(stderr, "p = %d < q = %d\n", p, q); exit(1);
  }
  for (i = a; i < p; i++) {
    if (A[i] > pivot) {
      fprintf(stderr, "A[%d] = %d > pivot = %d\n", i, A[i], pivot);
    }
  }
  for (i = b; i > q; i--) {
    if (A[i] < pivot) {
      fprintf(stderr, "A[%d] = %d < pivot = %d\n", i, A[i], pivot);
    }
  }
}

void check_sorted(elem_type * A, int a, int b)
{
  int i;
  for (i = a; i <= b - 1; i++) {
    if (A[i] > A[i + 1]) {
      fprintf(stderr, "bomb a[%d] = %d > a[%d] = %d\n", 
	      i, A[i], i + 1, A[i + 1]);
      exit(1);
    }
  }
}

/* sort A[a]...A[b] (inclusive) */

void insertion_sort(elem_type * A, int a, int c)
{
  int i;

  for (i = a; i <= c; i++) {
    /* find MIN element in A[i] .. A[c] */
    elem_type t;
    int m = i;
    int min = A[i];
    int j;
    for (j = i + 1; j <= c; j++) {
      if (A[j] < min) { m = j; min = A[j]; }
    }
    /* swap A[i] and A[m] */
    t = A[i];
    A[i] = A[m];
    A[m] = t;
  }
}

int partition_array(elem_type * A, int a, int c)
{
  int b = (a + c)/2;
  elem_type p1 = A[a], p2 = A[b], p3 = A[c];
  elem_type pivot;
  int p, q;
  
#if PARALLEL
  ST_POLLING();
#endif
  
  if (p1 <= p2) { 
    if (p3 <= p1) {		/* p3 <= p1 <= p2 */
      A[a] = p3; A[b] = p1; A[c] = p2;
      pivot = p1;
    } else if (p3 <= p2) {	/* p1 < p3 <= p2 */
      A[b] = p3; A[c] = p2;
      pivot = p3;
    } else {			/* p1 <= p2 < p3 */
      /* do nothing */
      pivot = p2;
    }
  } else { 
    if (p3 <= p2) {		/* p3 <= p2 < p1 */
      A[a] = p3; A[c] = p1;
      pivot = p2;
    } else if (p3 <= p1) {	/* p2 <= p3 <= p1 */
      A[a] = p2; A[b] = p3; A[c] = p1;
      pivot = p3;
    } else {			/* p2 < p1 < p3 */
      A[a] = p2; A[b] = p1;
      pivot = p1;
    }
  }
  
  p = a + 1; q = c - 1;
  while (p <= q) {
    /* safe because A[c] >= pivot */
    while(A[p] < pivot) p++;
    /* safe because A[a] <= pivot */
    while(A[q] > pivot) q--;
    /* here A[p] >= pivot and A[q] =< pivot */
    if (p <= q) {
      elem_type t = A[p];
      A[p] = A[q];
      A[q] = t;
      p++; q--;
    }
#if PARALLEL
    ST_POLLING();
#endif
  }
  /* now we have:
     (1) p >= q + 1, (2) for each i < p, A[i] <= pivot, and
     (3) for each i > q, A[i] >= pivot */
  return p;
}

typedef unsigned char uchar;
void qs1(elem_type * A, int a, int c
#if PARALLEL
	 , st_join_counter_t *jc
#endif
	 )
{
  /* printf("qs1: A[%d] ... A[%d]\n", a, c); */
  if (c - a < 4) {
    insertion_sort(A, a, c);
#if PARALLEL
    st_join_counter_finish(jc);
#endif
  } else {
    int p = partition_array(A, a, c);
#if !PARALLEL			/* serial */
    {
      qs1(A, a, p - 1);
      qs1(A, p, c);
    }
#else  /* PARALLEL */
    {				/* parallel */
      st_join_counter_t j[1];	/* synchronization counter */
      /* initialize it to have two threads to join */
      st_join_counter_init(j, 2);
      /* parallel recursion */
      ST_THREAD_CREATE(qs1(A, a, p - 1, j));
      qs1(A, p, c, j);
      /* wait for completion */
      st_join_counter_wait(j);
      ST_POLLING();
      st_join_counter_finish(jc);
    }
#endif
  }
}

void qs(elem_type * A, int n)
{
#if PARALLEL
  st_join_counter_t jc[1];
  st_join_counter_init(jc, 1);
#endif
  qs1(A, 0, n - 1
#if PARALLEL
      , jc
#endif
      );
#if PARALLEL
  st_join_counter_wait(jc);
#endif
}

void gen_array(elem_type * A, int n)
{
  int i;
  for (i = 0; i < n; i++) {
    A[i] = (elem_type)((i + 1) * 1234567  % (10 * n));
  }
}

int st_main(int argc, char ** argv)
{
  if (argc != 2) {
    fprintf(stderr, "usage: %s N\n", argv[0]);
    st_wg_exit((void *)1);
  } else {
    int n = atoi(argv[1]);
    long t0, t1;
    elem_type * A = (elem_type *)MM_ALLOC(sizeof(elem_type) * n);
    gen_array(A, n);
    /* uncomment this to demonstrate profiler */
    /* st_begin_profile(); */
    t0 = st_current_time_ms();
    qs(A, n);
    t1 = st_current_time_ms();
    /* uncomment this to demonstrate profiler */
    /* st_end_profile(); */
    check_sorted(A, 0, n - 1);
    printf("sorted %d elements in %ld ms\n", n, t1 - t0);
  }
  return 0;
}

