/*
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, October 16, 1998 */

# include "gc_priv.h"

# include <stdio.h>
# ifndef MACOS
#   include <signal.h>
#   include <sys/types.h>
# endif

static word min_words_allocd GC_PROTO((void)); 

/*
 * Separate free lists are maintained for different sized objects
 * up to MAXOBJSZ.
 * The call GC_allocobj(SGC_MYENV_ARG1 i,k) ensures that the freelist for
 * kind k objects of size i points to a non-empty
 * free list. It returns a pointer to the first entry on the free list.
 * In a single-threaded world, GC_allocobj may be called to allocate
 * an object of (small) size i as follows:
 *
 *            opp = &(GC_objfreelist[i]);
 *            if (*opp == 0) GC_allocobj(SGC_MYENV_ARG1 i, NORMAL);
 *            ptr = *opp;
 *            *opp = obj_link(ptr);
 *
 * Note that this is very fast if the free list is non-empty; it should
 * only involve the execution of 4 or 5 simple instructions.
 * All composite objects on freelists are cleared, except for
 * their first word.
 */

/*
 *  The allocator uses GC_allochblk to allocate large chunks of objects.
 * These chunks all start on addresses which are multiples of
 * HBLKSZ.   Each allocated chunk has an associated header,
 * which can be located quickly based on the address of the chunk.
 * (See headers.c for details.) 
 * This makes it possible to check quickly whether an
 * arbitrary address corresponds to an object administered by the
 * allocator.
 */

char * GC_copyright[] =
{"Copyright 1988,1989 Hans-J. Boehm and Alan J. Demers",
"Copyright (c) 1991-1995 by Xerox Corporation.  All rights reserved.",
"Copyright (c) 1998 by Kenjiro Taura, Toshio Endo. All rights reserved.",
"Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.",
"Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.",
"THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY",
" EXPRESSED OR IMPLIED.  ANY USE IS AT YOUR OWN RISK."};

struct _GC_publicinfo *GC_publicinfop = NULL;

/* some more variables */

int GC_incremental = 0;
    /* parallel version does not support incremental GC */
    /* and GC_incremental is always 0. */

int GC_never_stop_func GC_PROTO((void)) { return(0); }

CLOCK_TYPE GC_start_time;

int GC_timeout_stop_func GC_PROTO((void))
{
    CLOCK_TYPE current_time;
    static unsigned count = 0;
    unsigned long time_diff;
    
    if ((count++ & 3) != 0) return(0);
    GET_TIME(current_time);
    time_diff = MS_TIME_DIFF(current_time,GC_start_time);
    if (time_diff >= TIME_LIMIT) {
#   	ifdef PRINTSTATS
	    GC_printf0("Abandoning stopped marking after ");
	    GC_printf1("%lu msecs\n", (unsigned long)time_diff);
#	endif
    	return(1);
    }
    return(0);
}

/* Return the minimum number of words that must be allocated between	*/
/* collections to amortize the collection cost.				*/
static word min_words_allocd()
{
#   ifndef THREADS
    int dummy;
    register signed_word stack_size = (ptr_t)(&dummy) - GC_stackbottom;
#   else
    /* We punt, for now. */
    register signed_word stack_size = 10000;
#   endif
    register word total_root_size;  /* includes double stack size,	*/
    				    /* since the stack is expensive	*/
    				    /* to scan.				*/
    
    if (stack_size < 0) stack_size = -stack_size;
    total_root_size = 2 * stack_size + GC_root_size;
    if (GC_incremental) {
        return(BYTES_TO_WORDS(GC_heapsize + total_root_size)
               / (2 * GC_free_space_divisor));
    } else {
        return(BYTES_TO_WORDS(GC_heapsize + total_root_size)
               / GC_free_space_divisor);
    }
}

/* Return the number of words allocated, adjusted for explicit storage	*/
/* management, etc..  This number is used in deciding when to trigger	*/
/* collections.								*/
word GC_adj_words_allocd()
{
    register signed_word result;
    register signed_word expl_managed =
    		BYTES_TO_WORDS((long)GC_sum_app_stat.non_gc_bytes
    				- (long)GC_sum_app_stat.non_gc_bytes_at_gc);
    
    /* Don't count what was explicitly freed, or newly allocated for	*/
    /* explicit management.  Note that deallocating an explicitly	*/
    /* managed object should not alter result, assuming the client	*/
    /* is playing by the rules.						*/
    result = (signed_word)GC_sum_app_stat.words_allocd
    	     - (signed_word)GC_sum_app_stat.mem_freed - expl_managed;
    if (result > (signed_word)GC_sum_app_stat.words_allocd) {
        result = GC_sum_app_stat.words_allocd;
    	/* probably client bug or unfortunate scheduling */
    }
    result += GC_sum_app_stat.words_finalized;
    	/* We count objects enqueued for finalization as though they	*/
    	/* had been reallocated this round. Finalization is user	*/
    	/* visible progress.  And if we don't count this, we have	*/
    	/* stability problems for programs that finalize all objects.	*/
    result += GC_sum_app_stat.words_wasted;
     	/* This doesn't reflect useful work.  But if there is lots of	*/
     	/* new fragmentation, the same is probably true of the heap,	*/
     	/* and the collection will be correspondingly cheaper.		*/
    if (result < (signed_word)(GC_sum_app_stat.words_allocd >> 3)) {
    	/* Always count at least 1/8 of the allocations.  We don't want	*/
    	/* to collect too infrequently, since that would inhibit	*/
    	/* coalescing of free storage blocks.				*/
    	/* This also makes us partially robust against client bugs.	*/
        return(GC_sum_app_stat.words_allocd >> 3);
    } else {
        return(result);
    }
}


/* Clear up a few frames worth of garbage left at the top of the stack.	*/
/* This is used to prevent us from accidentally treating garbade left	*/
/* on the stack by other parts of the collector as roots.  This 	*/
/* differs from the code in misc.c, which actually tries to keep the	*/
/* stack clear of long-lived, client-generated garbage.			*/
void GC_clear_a_few_frames()
{
#   define NWORDS 64
    word frames[NWORDS];
    register int i;
    
    for (i = 0; i < NWORDS; i++) frames[i] = 0;
}

/* Have we allocated enough to amortize a collection? */
GC_bool GC_should_collect()
{
    return(GC_adj_words_allocd() >= min_words_allocd());
}

/*
 * Perform n units of garbage collection work.  A unit is intended to touch
 * roughly a GC_RATE pages.  Every once in a while, we do more than that.
 */
# define GC_RATE 8

int GC_deficit = 0;	/* The number of extra calls to GC_mark_some	*/
			/* that we have made.				*/
			/* Negative values are equivalent to 0.		*/

/*
 * Use the chunk of memory starting at p of syze bytes as part of the heap.
 * Assumes p is HBLKSIZE aligned, and bytes is a multiple of HBLKSIZE.
 */
void GC_add_to_heap(SGC_MYENV_ARG1 p, bytes)
SGC_MYENV_ARG_DECL2
struct hblk *p;
word bytes;
{
    word words;

    PSM_mutex_lock(&GC_heap_sects_ml);
    if (GC_n_heap_sects >= MAX_HEAP_SECTS) {
    	ABORT("Too many heap sections: Increase MAXHINCR or MAX_HEAP_SECTS");
    }
    if (!GC_install_header(SGC_MYENV_ARG1 p)) {
    	/* This is extremely unlikely. Can't add it.  This will		*/
    	/* almost certainly result in a	0 return from the allocator,	*/
    	/* which is entirely appropriate.				*/
	PSM_mutex_unlock(&GC_heap_sects_ml);
    	return;
    }
    GC_heap_sects[GC_n_heap_sects].hs_start = (ptr_t)p;
    GC_heap_sects[GC_n_heap_sects].hs_bytes = bytes;
    GC_n_heap_sects++;
    words = BYTES_TO_WORDS(bytes - HDR_BYTES);
    HDR(p) -> hb_sz = words;
    GC_freehblk(p);
    GC_heapsize += bytes;
    if ((ptr_t)p <= GC_least_plausible_heap_addr
        || GC_least_plausible_heap_addr == 0) {
        GC_least_plausible_heap_addr = (ptr_t)p - sizeof(word);
        	/* Making it a little smaller than necessary prevents	*/
        	/* us from getting a false hit from the variable	*/
        	/* itself.  There's some unintentional reflection	*/
        	/* here.						*/
    }
    if ((ptr_t)p + bytes >= GC_greatest_plausible_heap_addr) {
        GC_greatest_plausible_heap_addr = (ptr_t)p + bytes;
    }
    PSM_mutex_unlock(&GC_heap_sects_ml);
}

# if !defined(NO_DEBUGGING)
void GC_print_heap_sects()
{
    register unsigned i;
    
    GC_printf1("Total heap size: %lu\n", (unsigned long) GC_heapsize);
    for (i = 0; i < GC_n_heap_sects; i++) {
        unsigned long start = (unsigned long) GC_heap_sects[i].hs_start;
        unsigned long len = (unsigned long) GC_heap_sects[i].hs_bytes;
        struct hblk *h;
        unsigned nbl = 0;
        
    	GC_printf3("Section %ld from 0x%lx to 0x%lx ", (unsigned long)i,
    		   start, (unsigned long)(start + len));
    	for (h = (struct hblk *)start; h < (struct hblk *)(start + len); h++) {
    	    if (GC_is_black_listed(h, HBLKSIZE)) nbl++;
    	}
    	GC_printf2("%lu/%lu blacklisted\n", (unsigned long)nbl,
    		   (unsigned long)(len/HBLKSIZE));
    }
}
# endif

ptr_t GC_max(x,y)
ptr_t x, y;
{
    return(x > y? x : y);
}

ptr_t GC_min(x,y)
ptr_t x, y;
{
    return(x < y? x : y);
}

# if defined(__STDC__) || defined(__cplusplus)
    void GC_set_max_heap_size(GC_word n)
# else
    void GC_set_max_heap_size(n)
    GC_word n;
# endif
{
    GC_max_heapsize = n;
}

/*
 * this explicitly increases the size of the heap.  It is used
 * internally, but may also be invoked from GC_expand_hp by the user.
 * The argument is in units of HBLKSIZE.
 * Tiny values of n are rounded up.
 * Returns FALSE on failure.
 */
GC_bool GC_expand_hp_inner(SGC_MYENV_ARG1 n)
SGC_MYENV_ARG_DECL2
word n;
{
    word bytes;
    struct hblk * space;
    word expansion_slop;	/* Number of bytes by which we expect the */
    				/* heap to expand soon.			  */

    if (n < MINHINCR) n = MINHINCR;
    bytes = n * HBLKSIZE;
    
    if (GC_max_heapsize != 0 && GC_heapsize + bytes > GC_max_heapsize) {
        /* Exceeded self-imposed limit */
        return(FALSE);
    }
    space = GET_MEM(bytes);
    if( space == 0 ) {
	return(FALSE);
    }
#   ifdef PRINTSTATS
	GC_printf2("Increasing heap size by %lu after %lu allocated bytes\n",
	           (unsigned long)bytes,
	           (unsigned long)WORDS_TO_BYTES(GC_sum_app_stat.words_allocd));
/* Added by Endo */
	GC_printf3("Increasing heap %08x -- %08x (%08x)\n",
	           (unsigned long)space,
	           (unsigned long)space+bytes-1,
	           (unsigned long)bytes);
# 	ifdef UNDEFINED
	  GC_printf1("Root size = %lu\n", GC_root_size);
	  GC_print_block_list(); GC_print_hblkfreelist();
	  GC_printf0("\n");
#	endif
#   endif
    expansion_slop = 8 * WORDS_TO_BYTES(min_words_allocd());
    if (5 * HBLKSIZE * MAXHINCR > expansion_slop) {
        expansion_slop = 5 * HBLKSIZE * MAXHINCR;
    }
    if ((GC_last_heap_addr == 0 && !((word)space & SIGNB)) ||
        (GC_last_heap_addr != 0 && GC_last_heap_addr < (ptr_t)space)) {
        /* Assume the heap is growing up */
        GC_greatest_plausible_heap_addr =
            GC_max(GC_greatest_plausible_heap_addr,
                   (ptr_t)space + bytes + expansion_slop);
    } else {
        /* Heap is growing down */
        GC_least_plausible_heap_addr =
            GC_min(GC_least_plausible_heap_addr,
                   (ptr_t)space - expansion_slop);
    }
    GC_prev_heap_addr = GC_last_heap_addr;
    GC_last_heap_addr = (ptr_t)space;
    GC_add_to_heap(SGC_MYENV_ARG1 space, bytes);
    return(TRUE);
}

/* Really returns a bool, but it's externally visible, so that's clumsy. */
/* Arguments is in bytes.						*/
# if defined(__STDC__) || defined(__cplusplus)
  int GC_expand_hp(size_t bytes)
# else
  int GC_expand_hp(bytes)
  size_t bytes;
# endif
{
    DECLARE_MYENV3
    int result;

    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_alloc_ml, SGC_polling_inner);
#if 0
    if (!GC_is_initialized) GC_init_inner();
#endif
    result = (int)GC_expand_hp_inner(SGC_MYENV_ARG1 divHBLKSZ((word)bytes));
    PSM_mutex_unlock(&GC_g_alloc_ml);

    return(result);
}

void GC_gcollect GC_PROTO(())
{
    DECLARE_MYENV3
    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_alloc_ml, SGC_polling_inner);
    SGC_gcollect_inner(SGC_MYENV_ARG0);
    PSM_mutex_unlock(&GC_g_alloc_ml);
}

GC_bool GC_collect_or_expand_inner GC_PROTO((SGC_MYENV_ARG_DECL1
					  word needed_blocks,
					  GC_bool ignore_off_page));

GC_bool GC_collect_or_expand(SGC_MYENV_ARG1 needed_blocks, ignore_off_page)
SGC_MYENV_ARG_DECL2
word needed_blocks;
GC_bool ignore_off_page;
{
    GC_bool res;
    word oldheapsize = GC_heapsize;
    GC_word oldgcno = GC_gc_no;

    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_g_alloc_ml, SGC_polling_inner);
    if (oldheapsize != GC_heapsize || oldgcno != GC_gc_no) {
#if 0
	GC_err_printf0("collect_or_expand: may be improved by other thread.\n");
	GC_err_printf4("collect_or_expand: heapsize %dk->%dk, gc_no %d->%d\n",
		       oldheapsize/1024, GC_heapsize/1024,
		       oldgcno, GC_gc_no);
#endif
	PSM_mutex_unlock(&GC_g_alloc_ml);
	return TRUE;
    }
    res = GC_collect_or_expand_inner(SGC_MYENV_ARG1 needed_blocks, ignore_off_page);
    PSM_mutex_unlock(&GC_g_alloc_ml);
    return res;
}

GC_bool GC_collect_or_expand_inner(SGC_MYENV_ARG1 needed_blocks, ignore_off_page)
SGC_MYENV_ARG_DECL2
word needed_blocks;
GC_bool ignore_off_page;
{
    GC_sumup_app_counters();

    if (!GC_incremental && !GC_dont_gc && GC_should_collect()) {
	SGC_gcollect_inner(SGC_MYENV_ARG0);
    } else {
      word blocks_to_get = GC_heapsize/(HBLKSIZE*GC_free_space_divisor)
      			   + needed_blocks;
      if (blocks_to_get > MAXHINCR) {
          word slop;
          
          if (ignore_off_page) {
              slop = 4;
          } else {
	      slop = 2*divHBLKSZ(BL_LIMIT);
	      if (slop > needed_blocks) slop = needed_blocks;
	  }
          if (needed_blocks + slop > MAXHINCR) {
              blocks_to_get = needed_blocks + slop;
          } else {
              blocks_to_get = MAXHINCR;
          }
      }
      if (!GC_expand_hp_inner(SGC_MYENV_ARG1 blocks_to_get)
        && !GC_expand_hp_inner(SGC_MYENV_ARG1 needed_blocks)) {
      	if (GC_fail_count++ < 10) {
      	    WARN("Out of Memory!  Trying to continue ...\n", 0);
	    SGC_gcollect_inner(SGC_MYENV_ARG0);
	} else {
	    WARN("Out of Memory!  Returning NIL!\n", 0);
	    return(FALSE);
	}
      } else if (GC_fail_count) {
#	  ifdef PRINTSTATS
	    GC_printf0("Memory available again ...\n");
#	  endif
	  GC_fail_count = 0;
      }
    }
    return(TRUE);
}

/*
 * Make sure the object free list for sz is not empty.
 * Return a pointer to the first object on the free list.
 * The object MUST BE REMOVED FROM THE FREE LIST BY THE CALLER.
 * Assumes we hold the allocator lock and signals are disabled.
 *
 */
ptr_t GC_allocobj(SGC_MYENV_ARG1 sz, kind)
SGC_MYENV_ARG_DECL2
     word sz;
     int kind;
{
    DECLARE_MYENV
    register ptr_t * flh = &GC_mya_freelist[kind][sz];
    
    if (sz == 0) return(0);
    while (1) {
	if (*flh) {
	    break;
	}
#if 0
	/* Do our share of marking work */
        if(GC_incremental && !GC_dont_gc) GC_collect_a_little_inner(1);
#endif
	/* Sweep blocks for objects of this size */
	GC_continue_reclaim(SGC_MYENV_ARG1 sz, kind);
	if (*flh) {
	    break;
	}
	GC_continue_reclaim_g(SGC_MYENV_ARG1 sz, kind);
	if (*flh) {
	    break;
	}
	PSM_mutex_unlock(&GC_mya_alloc_ml); /* Unlock to avoid deadlock */
	GC_new_hblk(SGC_MYENV_ARG1 sz, kind);
	PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_mya_alloc_ml, SGC_polling_inner);
	if (*flh) {
	    break;
	}
	PSM_mutex_unlock(&GC_mya_alloc_ml); /* Unlock to avoid deadlock */
	if (!GC_collect_or_expand(SGC_MYENV_ARG1 (word)1,FALSE)) {
	    PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_mya_alloc_ml, SGC_polling_inner);
	    return(0);
	}
	PSM_mutex_lock_polling_inner(SGC_MYENV_ARG1 &GC_mya_alloc_ml, SGC_polling_inner);
/*fprintf(stderr, "[allocobj]loop...\n");*/
    }
    return(*flh);
}
