/* 
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991, 1992 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Endo, October 16, 1998 */
  
/* Routines for maintaining maps describing heap block
 * layouts for various object sizes.  Allows fast pointer validity checks
 * and fast location of object start locations on machines (such as SPARC)
 * with slow division.
 */
 
# include "gc_priv.h"

void GC_obj_map_init()
{
    register int displ;

    /* Allocate invalid_map */
    GC_invalid_map = GC_scratch_g_alloc(MAP_SIZE);
    if (GC_invalid_map == 0) {
	GC_err_printf0("Cant initialize GC_invalid_map: insufficient memory\n");
	EXIT();
    }
    for (displ = 0; displ < HBLKSIZE; displ++) {
	MAP_ENTRY(GC_invalid_map, displ) = OBJ_INVALID;
    }

    /* Preallocate large object map.  It's otherwise inconvenient to 	*/
    /* deal with failure.						*/
    if (!GC_add_map_entry((word)0)) {
        GC_err_printf0("Can't start up: not enough memory\n");
        EXIT();
    }

    PSM_mutex_init(&GC_obj_map_ml);
}

/* Invalidate the object map associated with a block.	Free blocks	*/
/* are identified by invalid maps.					*/
void GC_invalidate_map(hhdr)
hdr *hhdr;
{
/*    register int displ; */
    
    if (GC_invalid_map == 0) {
	fprintf(stderr, "ERROR: obj_map is not initialized..\n");
    }
    hhdr -> hb_map = GC_invalid_map;
}

/* Consider pointers that are offset bytes displaced from the beginning */
/* of an object to be valid.                                            */

# if defined(__STDC__) || defined(__cplusplus)
    void GC_register_displacement(GC_word offset)
# else
    void GC_register_displacement(offset) 
    GC_word offset;
# endif
{
# ifndef ALL_INTERIOR_POINTERS
    DISABLE_SIGNALS();
    PSM_mutex_lock_polling(&GC_g_alloc_ml, SGC_polling);
    GC_register_displacement_inner(offset);
    PSM_mutex_unlock(&GC_g_alloc_ml);
    ENABLE_SIGNALS();
# endif
}

void GC_register_displacement_inner(offset) 
word offset;
{
# ifndef ALL_INTERIOR_POINTERS
    register unsigned i;
    
    if (offset > MAX_OFFSET) {
        ABORT("Bad argument to GC_register_displacement");
    }
    if (!GC_valid_offsets[offset]) {
      GC_valid_offsets[offset] = TRUE;
      GC_modws_valid_offsets[offset % sizeof(word)] = TRUE;
      for (i = 0; i <= MAXOBJSZ; i++) {
          if (GC_obj_map[i] != 0) {
             if (i == 0) {
               GC_obj_map[i][offset + HDR_BYTES] = (char)BYTES_TO_WORDS(offset);
             } else {
               register unsigned j;
               register unsigned lb = WORDS_TO_BYTES(i);
               
               if (offset < lb) {
                 for (j = offset + HDR_BYTES; j < HBLKSIZE; j += lb) {
                   GC_obj_map[i][j] = (char)BYTES_TO_WORDS(offset);
                 }
               }
             }
          }
      }
    }
# endif
}


/* Add a heap block map for objects of size sz to obj_map.	*/
/* Return FALSE on failure.					*/
GC_bool GC_add_map_entry(sz)
word sz;
{
    register unsigned obj_start;
    register unsigned displ;
    register char * new_map;
    
    if (sz > MAXOBJSZ) sz = 0;

    /* Check before lock. This is for performance. */
    if (GC_obj_map[sz] != 0) {
        return(TRUE);
    }

    PSM_mutex_lock(&GC_obj_map_ml);

    /* Check after lock */
    if (GC_obj_map[sz] != 0) {
	PSM_mutex_unlock(&GC_obj_map_ml);
        return(TRUE);
    }

    new_map = GC_scratch_g_alloc(MAP_SIZE);
    if (new_map == 0) {
	PSM_mutex_unlock(&GC_obj_map_ml);
	return(FALSE);
    }
#   ifdef PRINTSTATS
        GC_printf1("Adding block map for size %lu\n", (unsigned long)sz);
#   endif
    for (displ = 0; displ < HBLKSIZE; displ++) {
        MAP_ENTRY(new_map,displ) = OBJ_INVALID;
    }
    if (sz == 0) {
        for(displ = 0; displ <= MAX_OFFSET; displ++) {
            if (OFFSET_VALID(displ)) {
                MAP_ENTRY(new_map,displ+HDR_BYTES) = BYTES_TO_WORDS(displ);
            }
        }
    } else {
        for (obj_start = HDR_BYTES;
             obj_start + WORDS_TO_BYTES(sz) <= HBLKSIZE;
             obj_start += WORDS_TO_BYTES(sz)) {
             for (displ = 0; displ < WORDS_TO_BYTES(sz); displ++) {
                 if (OFFSET_VALID(displ)) {
                     MAP_ENTRY(new_map, obj_start + displ) =
                     				BYTES_TO_WORDS(displ);
                 }
             }
        }
    }
    GC_obj_map[sz] = new_map;
    PSM_mutex_unlock(&GC_obj_map_ml);
    return(TRUE);
}
