/*
 * Copyright (c) 1998 by Toshio Endo. All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */

/* Some primitive (low-level) operations for Shared-Memory */

#ifndef PSM_H
#define PSM_H

#ifdef	__cplusplus
extern "C" {
#endif

#ifndef INLINE
#ifdef __GNUC__
    /*#define INLINE inline*/
#define INLINE 
#else
#define INLINE
#endif
#endif

#if !defined __GNUC__ && defined PSM_STANDALONE
#undef PSM_STANDALONE
#endif

#ifdef PSM_STANDALONE
/******************** Stand alone mode. You have to use GNU C compiler. */
/********************************** A little faster than standard mode. */

#if defined sparc || defined __sparc
#define PSM_SPARC
#endif

#if defined mips || defined __mips
#define PSM_MIPS
#endif

#if defined i386
#define PSM_I386
#endif

#if !defined PSM_SPARC && !defined PSM_MIPS && !defined PSM_I386
#error unknown architecture!
#endif

#if !defined __GNUC__
#error GCC only!
#endif

#ifdef PSM_SPARC

/* This file requires SPARC-V8plus architecture. */
/* This file uses SWAP, CAS, MEMBAR(STBAR). */
/* Use -Wa,-xarch=v8plus (or -Wa,-xarch=v8plusa) option. */

#define PSM_SPARC_SWAP(p, n) \
    asm volatile("swap [%1], %0" : "=r" (n) : "p" (p), "0" (n))
#define PSM_SPARC_CAS(p, rs, rd) \
    asm volatile("cas [%1], %2, %0" : "=r" (rd) : "p" (p), "r" (rs), "0" (rd))
#define PSM_SPARC_MEMBAR() \
    asm volatile("membar #StoreLoad | #Sync" : : )

INLINE static void PSM_sync_write_c(long volatile *p, long n)
{
    *p = n;
    PSM_SPARC_MEMBAR();
    return;
}

INLINE static int PSM_compare_and_swap_c(long volatile *p, long oldn, long newn)
{
    PSM_SPARC_CAS(p, oldn, newn);
    return (oldn == newn);
}

#elif defined PSM_MIPS

/* This file requires MIPS-II architecture. */
/* This file uses LL, SC, SYNC. */
/* Use -mips2 option. */

#define PSM_MIPS_LL(n, p) asm volatile("ll %0, (%1)" : "=r" (n) : "p" (p))
#define PSM_MIPS_SC(n, p) \
    asm volatile("sc %0, (%1)" : "=r" (n) : "p" (p), "0" (n))
#define PSM_MIPS_SYNC() asm volatile("sync" : : )

INLINE static void PSM_sync_write_c(long volatile *p, long n)
{
    *p = n;
    PSM_MIPS_SYNC();
    return;
}

#if 1
static int PSM_compare_and_swap_c(long volatile *p, long oldn, long newn)
{
    /* 1st argument is $4 */
    /* 2nd argument is $5 */
    /* 3rd argument is $6 */
    /* Return value is $2 */
    asm volatile(
		 "ll	$2,($4)\n"
		 ".set	noreorder\n"
		 ".set	nomacro\n"
		 "bne	$2,$5,PSM_CAS_C1\n"
		 "move	$2,$0\n"
		 ".set	macro\n"
		 ".set	reorder\n"
		 "sc	$6,($4)\n"
		 "sltu	$2,$0,$6\n"
	 "PSM_CAS_C1:\n"
    );
}
#else
static int PSM_compare_and_swap_c(long volatile *p, long oldn, long newn)
{
    register long data;
    register long newn_reg = newn;
    register long volatile *p_reg = p;
    register long oldn_reg = oldn;
    PSM_MIPS_LL(data, p_reg);
    if (data == oldn_reg) {
	PSM_MIPS_SC(newn_reg, p_reg);
	if (newn_reg) {
	    /* Succeeded */
	    return 1;
	}
    }
    return 0;
}
#endif

#elif defined PSM_I386

/* This file requires 486 architecture. */
/* This file uses XCHG, lock+CMPXCHG. */
#define PSM_INTEL_XCHG(p, n) \
    asm volatile("xchg (%1), %0" : "=r" (n) : "p" (p), "0" (n))
#define PSM_INTEL_XADD(p, n) \
     asm volatile("lock\n\txadd %0, (%1)": : "r" (n), "p" (p))

INLINE static void PSM_sync_write_c(long volatile *p, long n)
{
    PSM_INTEL_XCHG(p, n);
    return;
}

static int PSM_compare_and_swap_c(long volatile *p, long oldn, long newn)
{
    /* First argument is 8(%ebp) */
    /* Second argument is 12(%ebp) */
    /* Third argument is 16(%ebp) */
    /* Return value is %eax */

    asm volatile(
	"pushl %ebx\n"
	"pushl %ecx\n"
	"movl 8(%ebp),%ebx\n"
	"movl 12(%ebp),%eax\n"
	"movl 16(%ebp),%ecx\n"
	"lock\n"
	"cmpxchg %ecx,(%ebx)\n"
	"movl $0,%eax\n"
	"sete %al\n"
	"popl %ecx\n"
	"popl %ebx\n"
    );
}

#endif

#define PSM_SYNC_WRITE PSM_sync_write_c
#define PSM_COMPARE_AND_SWAP PSM_compare_and_swap_c

#else /* PSM_STANDALONE */
/********************** Standard mode. You have to link psm_xxx.s. */

/* Architecture dependent functions defined in psm_xxx.s */

void PSM_sync_write (long volatile *p, long n);
int PSM_compare_and_swap (long volatile *p, long oldn, long newn);

#define PSM_SYNC_WRITE PSM_sync_write
#define PSM_COMPARE_AND_SWAP PSM_compare_and_swap

#endif /* PSM_STANDALONE */

/******************************************* Type definition */

typedef long PSM_mutex_t;

typedef struct {
    long npe;
    long sense;
    long count;
} PSM_barrier_t;

/******************************************* mutual exclusion */

INLINE static void PSM_mutex_init(PSM_mutex_t *mutex_p)
{
    PSM_SYNC_WRITE(mutex_p, 0);
}

INLINE static void PSM_mutex_lock(PSM_mutex_t *mutex_p)
{
    long volatile *p;
    p = mutex_p;
    while (1) {
	if (*p) {
	    continue;
	}
	if (PSM_COMPARE_AND_SWAP(p, 0, 1)) {
	    break;
	}
    }
}

INLINE static int PSM_mutex_trylock(PSM_mutex_t *mutex_p)
{
    if (*mutex_p) {
	return 0;
    }
    return PSM_COMPARE_AND_SWAP(mutex_p, 0, 1);
}

INLINE static void PSM_mutex_lock_polling(PSM_mutex_t *mutex_p,
					  void (*polling_func)(void))
{
    while (PSM_mutex_trylock(mutex_p) == 0) {
	(*polling_func)();
    }
}

INLINE static void PSM_mutex_unlock(PSM_mutex_t *mutex_p)
{
    PSM_SYNC_WRITE(mutex_p, 0);
}

INLINE static int PSM_mutex_islocked(PSM_mutex_t *mutex_p)
{
    return (*mutex_p != 0);
}

INLINE static void PSM_mutex_destroy(PSM_mutex_t *mutex_p)
{
}

/*************************** atomic bit-set */
/* if ((*addr & mask) == 0) (*addr) |= mask */
INLINE static int PSM_test_and_set(unsigned long *addr, unsigned long mask)
{
    unsigned long volatile *p = addr;
    unsigned long data;
    unsigned long new_data;

    while (1) {
	data = *p;
	if (data & mask) {
	    /* Already set */
	    return 0;
	}
	else {
            new_data = data | mask;
            if (PSM_COMPARE_AND_SWAP((long volatile *)p, data, new_data)) {
		/* Succeeded */
		return 1;
	    }
	}
    }
}

/******************************** simple counter */
INLINE static void PSM_counter_add(long *addr, long inc)
{
    long volatile *p = addr;
    long count;
    long new_count;

    if (inc == 0) return;

    while (1) {
	count = *p;
	new_count = count + inc;
	if (PSM_COMPARE_AND_SWAP(p, count, new_count)) {
	    /* Succeeded */
	    return;
	}
    }
}

/*************************** barrier synchronization */

static void PSM_barrier_init(PSM_barrier_t *barrier_p, long npe)
{
    barrier_p->npe = npe;
    barrier_p->sense = 0;
    barrier_p->count = 0;
}

static void PSM_barrier_sync(PSM_barrier_t *barrier_p)
{
    long volatile *psense = &barrier_p->sense;
    long volatile *pcount = &barrier_p->count;
    long local_sense;
    long count;
    long new_count;

    local_sense = 1 - *psense;

    /* increment count */
    while (1) {
	count = *pcount;
	new_count = count + 1;
	if (PSM_COMPARE_AND_SWAP(pcount, count, new_count)) {
	    /* Succeeded */
	    break;
	}
    }

    if (new_count == barrier_p->npe) {
        *pcount = 0;
	PSM_SYNC_WRITE(psense, local_sense);
    }
    else {
        while (local_sense != *psense) ;
    }
}

static void PSM_barrier_sync_polling(PSM_barrier_t *barrier_p,
				     void (*polling_func)(void))
{
    long volatile *psense = &barrier_p->sense;
    long volatile *pcount = &barrier_p->count;
    long local_sense;
    long count;
    long new_count;

    local_sense = 1 - *psense;

    /* increment count */
    while (1) {
	count = *pcount;
	new_count = count + 1;
	if (PSM_COMPARE_AND_SWAP(pcount, count, new_count)) {
	    /* Succeeded */
	    break;
	}
	(*polling_func)();
    }

    if (new_count == barrier_p->npe) {
        *pcount = 0;
	PSM_SYNC_WRITE(psense, local_sense);
    }
    else {
        while (local_sense != *psense) {
	    (*polling_func)();
	}
    }
}

static void PSM_barrier_destroy(PSM_barrier_t *barrier_p)
{
}

#ifdef	__cplusplus
}
#endif

#endif /* PSM_H */
