#if 1
/*
 * rna.cc
 */

#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#if STHREADS && !MTCAMP
#include <st.h>
#endif

/* currently, we do not free anything. so UNDERLYING_MM_SGC_EXP_FREE
   and UNDERLYING_MM_MALLOC will leak (without RALLOC) */
#ifndef UNDERLYING_MM
#define UNDERLYING_MM UNDERLYING_MM_SGC_EXP_FREE
#endif
#ifndef RALLOC
#define RALLOC 1
#endif

#include <st_list.h>

typedef double real;

/* given (stack-data E a b c d) it holds that a < b < c < d
   and (a + x) and (d - x) is bonded by hydro-bonded for each x < b - a
          ____
         /    \
         \    /
        b |  |c
    ______|  |______
        a     d
*/

struct stack_datum : st_mm
{
  stack_datum (real E0, int a0, int b0, int c0, int d0) {
    E = E0; a = a0; b = b0; c = c0; d = d0;
  }
  real E;
  int a, b, c, d;
};
typedef stack_datum * stack_datum_t;

struct stack_data : st_mm
{
  stack_data() {
    n_islets = 0;
    islets = 0;
    stack_regions = 0;
    n_stack_regions = 0;
  }
  int geometically_compatible_p(int, int);
  int geometically_compatible_all_p(int, st_list_t(int));
  int geometically_incompatible_all_p(int, st_list_t(int));
  void read_stack_regions(char *, int);
  void find_cliques();
  real ideal_energy(int);
  real initial_max_estimate();
  real n_possible_combinations();
  void print_islets();

  int n_islets;
  st_list_t(int) * islets;
  stack_datum_t * stack_regions;
  int n_stack_regions;
};
typedef stack_data * stack_data_t;

/* check if stack region X and Y are geometically compatible */

int stack_data::geometically_compatible_p (int x, int y)
{
  stack_datum_t xx = stack_regions[x];
  stack_datum_t yy = stack_regions[y];
  if (yy->d < xx->a) return 1;
  if (xx->d < yy->a) return 1;
  if (xx->b < yy->a && yy->d < xx->c) return 1;
  if (yy->b < xx->a && xx->d < yy->c) return 1;
  return 0;
}

/* true if X is compatible to any Y in YS */

int stack_data::geometically_compatible_all_p(int x, st_list_t(int) ys)
{
  st_list_t(int) yp;
  for (yp = ys; yp; yp = yp->cdr) {
    if (geometically_compatible_p(x, yp->car) == 0)
      return 0;
  }
  return 1;
}

int stack_data::geometically_incompatible_all_p(int x, st_list_t(int) ys)
{
  st_list_t(int) yp;
  for (yp = ys; yp; yp = yp->cdr) {
    if (geometically_compatible_p(x, yp->car))
      return 0;
  }
  return 1;
}

/* read stack-data from file */

int consistent_data_p (real E, int a, int b, int c, int d)
{
  if (E > 0.0 && a < b && b < c && c < d && b - a == d - c)
    return 1;
  if (E == 0.0 && a == 0 && b == 0 && c == 0 && d == 0)
    return 1;
  return 0;
}
 
#define MAX_LINE 127
stack_datum_t read_stack_datum (FILE * f)
{
  real E;
  int a, b, c, d;
  char s[MAX_LINE + 1];
  fgets (s, MAX_LINE, f);
  sscanf (s, "%lf %d %d %d %d", &E, &a, &b, &c, &d);
  
  if (consistent_data_p (E, a, b, c, d)) {
    return new stack_datum (E, a, b, c, d);
  } else {
    printf ("inconsistent stack region data: "
	    "E = %f, a = %d, b = %d, c = %d, d = %d\n",
	    E, a, b, c, d);
    st_wg_exit((void *)1);
    return 0;
  }
}

/* read at most MAX_N lines from FILENAME to read region data.
   data is set to stack_regions and the number of read lines to
   n_stack_regions */
void stack_data::read_stack_regions(char * filename, int max_n)
{
  FILE * f = fopen (filename, "r");
  if (f == 0) { printf ("cannot open %s\n", filename); st_wg_exit((void *)1); }
  st_list_t(stack_datum_t) r = 0;
  st_list_t(stack_datum_t) p;
  int n; int i;
  
  for (n = 0; n < max_n; n++) {
    stack_datum_t d = read_stack_datum (f);
    if (d->a == 0) break;
    r = cons (d, r);
  }
  printf ("%d stack regions have been read\n", n);
  stack_datum_t * v = new MM_DEFAULT_OBJ_PLACEMENT stack_datum_t [n];
  MM_ZERO_CLEAR((void *)v, n * sizeof(stack_datum_t));
  for (p = r, i = n; p; p = p->cdr, i--) {
    v[i - 1] = p->car;
  }
  /* set instance variables */
  stack_regions = v;
  n_stack_regions = max_n;
}

/* calc incompatible islets from stack regions */

void reorder_cliques_aux (st_list_t(int) * cliques, int b, int e)
{
  if (e >= b + 1) {
    int c = (b + e + 1) / 2;
    int piv = length (cliques[c]);
    int p = b; int q = e;
    while (p < q) {
      while (length (cliques[p]) < piv) p++;
      while (length (cliques[q]) > piv) q--;
      if (p <= q) {
	st_list_t(int) s = cliques[p];
	st_list_t(int) t = cliques[q];
	cliques[q] = s;
	cliques[p] = t;
	p++; q--;
      }
    }
    reorder_cliques_aux (cliques, b, q);
    reorder_cliques_aux (cliques, p, e);
  }
}

st_list_t(int) * reorder_cliques(st_list_t(int) * cliques, int n)
{
  reorder_cliques_aux (cliques, 0, n - 1);
  return cliques;
}

/* given stack_regions and n_stack_regions have been set, 
   set islets and n_islets */
void stack_data::find_cliques()
{
  stack_datum_t * regions = stack_regions;
  int n_regions = n_stack_regions;
  char * selected = new MM_DEFAULT_OBJ_PLACEMENT char [n_regions];
  MM_ZERO_CLEAR((void *)selected, sizeof(char) * n_regions);
  st_list_t(st_list_t(int)) cliques = 0;
  int i, j;

  for (i = 0; i < n_regions; i++) {
    if (selected[i] == 0) {
      /* skip regions already in some clique */
      st_list_t(int) clique = 0;
      for (j = i; j < n_regions; j++) {
	if (selected[j] == 0 
	    && geometically_incompatible_all_p (j, clique)) {
	  selected[j] = 1; 
	  clique = cons (j, clique);
	} 
      }
      cliques = cons (clique, cliques);
    }
  }
  int l = length (cliques);
  st_list_t(int) * r = new MM_DEFAULT_OBJ_PLACEMENT st_list_t(int) [l];
  MM_ZERO_CLEAR((void *)r, sizeof(st_list_t(int)) * l);
  int n = 0;
  st_list_t(st_list_t(int)) p;
  for (p = cliques; p; p = p->cdr) {
    r[n] = reverse(p->car);
    n++;
  }
  
  /* set instance variables */
  islets = reorder_cliques(r, n);
  n_islets = n;
}

/* ideal  energy for Dth islet */
real stack_data::ideal_energy(int d)
{
  real x = stack_regions[islets[d]->car]->E;
  return x;
}

/* calculate maximum possible energy which would be obtained when we
   ignore geometical information other than islets */

real stack_data::initial_max_estimate()
{
  real s = 0.0;
  int i;
  for (i = 0; i < n_islets; i++) {
    s += stack_regions[islets[i]->car]->E;
  }
  return s;
}

real stack_data::n_possible_combinations()
{
  real p = 1.0;
  int i;
  for (i = 0; i < n_islets; i++) {
    p *= (real)(length (islets[i]) + 1);
  }
  return p;
}

/*  print-islets */

void print_int(int x)
{
  printf(" %d", x);
}

void stack_data::print_islets ()
{
  int i;
  for (i = 0; i < n_islets; i++) {
    st_list_for_each(print_int, islets[i]);
    printf ("\n");
  }
}

stack_data_t make_stack_data(char * filename, int max_n)
{
  stack_data * sd = new stack_data();
  sd->read_stack_regions(filename, max_n);
  sd->find_cliques();
  return sd;
}

/* data structure that describes an answer */

struct answer : st_mm
{
  answer (st_list_t(int) a, real e) { ans = a; E = e; }
  st_list_t(int) ans;
  real E;
};

#ifdef BLOCK_TEST
#define COUNT_NODES
#endif

/* manage search */
struct rna_manager : st_mm
{
  rna_manager(real am, real is, int bf) { 
    answers = 0; 
    ST_INT_LOC_INIT(&best_answer_lock, 0);
    best_answer = -1.0; 
    accept_margin = am;
    immediate_signal_threshold = is;
    ST_INT_LOC_INIT(&n_nodes_visited, 0);
    block_frequency = bf;
  }
  int should_block();
  void tick_n_nodes();
  int read_n_nodes();
  void update_best_answer(real);
  void add_answer(st_list_t(int), real, st_region_t);
  void evac_answers();
  real find_best_answer();
  real print_answers();
  int count_answers();

  /* answers found so far */
  st_list_t(answer*) answers;
  /* we accept answers which are largher than M - accept_margin where M is
     the best answer */
  real accept_margin;
  /* signal a new answer when we found an answer larger than 
     M + intermediate_signal_threshold even if we are in the middle 
     of search */
  real immediate_signal_threshold;
  
  /* lock for the following three variables */
  st_int_loc_t best_answer_lock;			
  /* monotonically increasing variables */
  real best_answer;		// best_answer found so far
  real abort_threshold;		// best_answer - accept_margin
  real signal_threshold;	// best_answer + immediate_signal_threshold
  st_int_loc_t n_nodes_visited;
  int block_frequency;
};

void rna_manager::tick_n_nodes()
{
  st_fetch_and_add_int(&n_nodes_visited, 1);
}

int rna_manager::read_n_nodes()
{
  return st_read_int(&n_nodes_visited);
}

int rna_manager::should_block()
{
  if ((read_n_nodes() + 1) % block_frequency == 0) return 1;
  else return 0;
}

void rna_manager::update_best_answer (real e)
{
  if (best_answer < e) {
    st_read_and_lock_int(&best_answer_lock);
    if (best_answer < e) {
      best_answer = e;
      abort_threshold = e - accept_margin;
      signal_threshold = immediate_signal_threshold + e;
    }
    st_write_and_unlock_int(&best_answer_lock, 0); // unlock
  }
}

void rna_manager::add_answer(st_list_t(int) taken, real e, st_region_t rg)
{
  answer * new_ans = new answer (taken, e);
  st_list_t(answer*) c = cons(new_ans, (st_list_t(answer*))0, rg);
  
  st_list_t(answer*) 
    as = (st_list_t(answer*))st_read_and_lock_ptr((st_ptr_loc_t *)&answers);
  c->cdr = as;
  st_write_and_unlock_ptr((st_ptr_loc_t *)&answers, (void *)c); // unlock
}

void rna_manager::evac_answers()
{
  answers = 0;
}

real rna_manager::find_best_answer()
{
  real r = 0.0;
  st_list_t(answer*) p;
  st_list_t(answer*) as = answers;
  for (p = as; p; p = p->cdr) {
    answer * a = p->car;
    if (a->E >= r) r = a->E;
  }
  return r;
}

real rna_manager::print_answers()
{
  st_list_t(answer *) p;
  st_list_t(answer*) as = answers;
  real b = find_best_answer();
  real acc = b - accept_margin;
  for (p = as; p; p = p->cdr) {
    answer * a = p->car;
    if (a->E >= acc) {
      printf ("answer = ");
      st_list_for_each(print_int, a->ans);
      printf (" E = %f\n", a->E);
    }
  }
  return b;
}

int rna_manager::count_answers()
{
  st_list_t(answer*) p;
  st_list_t(answer*) as = answers;
  int i = 0;
  for (p = as; p; p = p->cdr) {
    i++;
  }
  return i;
}

rna_manager * make_rna_manager(real am, real is, int bf)
{
  rna_manager * rm = new rna_manager(am, is, bf);
  rm->update_best_answer(0.0);
  return rm;
}

/*
   DEPTH: index of incompatible islet from which we select the next 
   stack region, which is equivalent to the depth in the search tree
   TAKEN: stack regions already taken
   MIN-E and MAX-E: minimum and maximum possible energies in the current
   configuration.  MIN-E is effectively summation of energies
   resulted by TAKEN.  MAX-E is initially the energy
   obtained by selecting the leftmost stack in each region, 
   ignoring other geometical information.  MAX-E is decremented at 
   each step when selecting a stack region but the leftmost one. */

void search_node (int depth, st_list_t(int) taken, real min_e, real max_e,
		  stack_data_t sd, rna_manager * rm, 
		  st_region_t rg
#if STHREADS && !MTCAMP
		  , st_join_counter_t * ack
#endif
		  )
{
  DECLARE_STACK_INV_CHECK;
#ifdef COUNT_NODES
  rm->tick_n_nodes();
#endif
#ifdef BLOCK_TEST
  if (rm->block_frequency && rm->should_block()) {
      st_temporal_suspend();
  }
#endif
  
  SAVE_STACK_INV();
  ST_POLLING();
  CHECK_STACK_INV();
  
  if (max_e < rm->abort_threshold) {
#if STHREADS && !MTCAMP
    st_join_counter_finish(ack);
#endif
  } else if (depth == sd->n_islets) {
    /* we reached a leaf, signal answer if it is a good one */
    if (rm->best_answer < min_e) rm->update_best_answer(min_e);
    rm->add_answer(taken, min_e, rg);
#if STHREADS && !MTCAMP
    st_join_counter_finish(ack);
#endif
  } else {
    /* tell other people if we significantly improve old value */
    if (rm->signal_threshold < min_e) {
      rm->update_best_answer(min_e);
    }
    /* we will hopefully reach an answer under this node,
       pick up stack regions from ISLETS[DEPTH] and recurse */
    st_list_t(int) s = sd->islets[depth];
    real ideal_e = sd->ideal_energy(depth);
    st_list_t(int) p;
    
	// create counter for the children (1 for the last node)
#if MTCAMP
    mtc_sync {
#elif STHREADS
      st_join_counter_t cack[1];
      st_join_counter_init(cack, 1);
#endif
      
      for (p = s; p; p = p->cdr) {
	int s = p->car;
	if (sd->geometically_compatible_all_p(s, taken)) {
	  real e = sd->stack_regions[s]->E;
	  SAVE_STACK_INV();
#if MTCAMP
	  mtc_fork {
	    search_node (depth + 1, cons(s, taken, rg), 
			 min_e + e, max_e - (ideal_e - e), 
			 sd, rm, rg);
	  }
#elif STHREADS
	  st_join_counter_spawn(cack);
	  ST_THREAD_CREATE(search_node(depth + 1, cons(s, taken, rg),
				       min_e + e, 
				       max_e - (ideal_e - e), sd, rm, rg,
				       cack));
#else  /* SEQUENTIAL */
	  search_node (depth + 1, cons(s, taken, rg), 
		       min_e + e, max_e - (ideal_e - e), sd, rm, rg);
#endif	
	  CHECK_STACK_INV();
	}
      }
      // src_add for this node has been done on creation
      SAVE_STACK_INV();
      search_node(depth + 1, taken, min_e, max_e - ideal_e, sd, rm, rg
#if STHREADS && !MTCAMP
		  , cack
#endif
		  );
      CHECK_STACK_INV();
      
      SAVE_STACK_INV();
      /* wait for completion */ 
#if MTCAMP
      ---;
#elif STHREADS
      st_join_counter_wait(cack);
#endif
      /* completion */ 
#if MTCAMP
    } /* mtc_sync */
#endif
    
    CHECK_STACK_INV();
    
	/* say I have done */
    SAVE_STACK_INV();
#if STHREADS && !MTCAMP
    st_join_counter_finish(ack);
#endif
    CHECK_STACK_INV();
    
    SAVE_STACK_INV();
    ST_POLLING();
    CHECK_STACK_INV();
  }
}

/*
 rna.sun FILE N IS AM PROFILE
 FILE --- stack data file
 N --- maximum elements to read from FILE
 IS --- threshould value above which we signal improvements
        to other processors even in the intermediate nodes
 AM --- threshould value within which we accept non-optimal
        solutions
*/

void rna_main (char * stack_file, int n_regions, real am, real is, int bf)
{
  stack_data_t sd = make_stack_data(stack_file, n_regions);
  rna_manager * rm = make_rna_manager(am, is, bf);
  real m = sd->initial_max_estimate();
  st_region_t rg = MM_MAKE_REGION_HIGHLY_CONCURRENT(10 * 1024 * 1024);
  int i;
    
  for (i = 0; i < 10; i++) {
    int t0, t1; real b;
#if STHREADS && !MTCAMP
    st_join_counter_t ack[1];
    st_join_counter_init(ack, 1);
#endif
    
    t0 = st_current_time_ms();
    
    search_node(0, 0, 0.0, m, sd, rm, rg
#if STHREADS && !MTCAMP
		, ack
#endif
		);
    t1 = st_current_time_ms();
    
#ifdef COUNT_NODES
    printf ("%d nodes visited out of %f possible combinations\n",
	    rm->read_n_nodes(), sd->n_possible_combinations());
#endif
    
    b = rm->print_answers();
    printf("best is %lf, accept_margin is %lf\n", b, rm->accept_margin);
    printf("the list holds %d answers\n", rm->count_answers());
    printf("search [ %d ... %d ] = %d msec\n", t0, t1, t1 - t0);
    
    rm->evac_answers();

    MM_RESET_REGION(rg);
    MM_GC_POINT();
  }

}

int st_main (int argc, char ** argv)
{
  DECLARE_STACK_INV_CHECK;
#ifdef BLOCK_TEST
  if (argc != 4) {
    printf ("usage: %s stack_file #_of_stack_regions block_frequency\n",
	    argv[0]);
    st_wg_exit(1);
  }
#else
  if (argc != 3) {
    printf ("usage: %s stack_file #_of_stack_regoins\n",
	    argv[0]);
    st_wg_exit((void *)1);
  }
#endif
  char * stack_file = argv[1];
  int n_regions = atoi (argv[2]);
  real is = 20.0;
  real am = 1.0;
#ifdef BLOCK_TEST
  int bf = atoi(argv[3]);
#else
  int bf = 0;
#endif
  
  MM_SETUP(/* GC_free_space_divisor = */ 4);
  
  SAVE_STACK_INV();
  rna_main(stack_file, n_regions, am, is, bf);
  CHECK_STACK_INV();

#if UNDERLYING_MM_IS_SGC
  printf ("%d GCs\n", GC_gc_no);
#endif

  return 0;
}

#else
#include <st.h>

int st_main (int argc, char ** argv)
{

}
#endif
